import os
from typing import Callable, Union
import time

from .parallelization import RDSProcessManager
from .utils import get_file_list, decompress_zstd


class Reader:
    """
    The core class to read the generated temporal wikipedia blocks.

    Basic usage:

        import twb
        reader = twb.Reader()  # Create an instance.
        reader.preload('./test/sample_data/minimal_sample.xml')  # Preload the files to be processed.

    You must preload the files to be processed before building the blocks.
    Preload function accepts a path to a file or a directory, and can be called multiple times (for multiple files).

    Attributes:
        files (list): A list of files to be read.
    """

    def __init__(self):
        self.files = []

    def preload(self, path: str):
        """
        Preload the files to be processed.
        It will not actually load to the memory until the build() method is called.
        :param path: the path of a file or a directory
        :raise ValueError: if the path is empty
        :raise FileNotFoundError: if the path does not exist
        """
        if not path:
            raise ValueError('The path cannot be empty.')
        if not os.path.exists(path):
            raise FileNotFoundError('The path does not exist.')
        self.files.extend(get_file_list(path))

    def decompress(self,
                   output_dir: str,
                   num_proc: Union[int, None] = None):
        """
        Decompress the files.
        :param output_dir: the output directory
        :param num_proc: the number of processes to use
        """

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        start_time = time.time()

        pm = RDSProcessManager(
            num_proc=num_proc
        )
        for file in self.files:
            pm.apply_async(
                executable=_decompress_file,
                args=(file, output_dir),
                use_controller=False
            )

        pm.close()
        pm.join()

        end_time = time.time()
        execution_duration = end_time - start_time
        print(f'[Read] Decompression finished in {execution_duration:.2f} seconds.')

    def map(self, func: Callable[[dict], None]):
        """
        Map a function to each block.
        :param func: the function to be mapped
        """
        # TODO: Implement this method.
        pass


def _decompress_file(path: str, output_dir: str):
    original_name = os.path.split(path)[1]
    decompressed_name = original_name.replace('.zst', '')
    decompressed_path = os.path.join(output_dir, decompressed_name)
    decompress_zstd(path, decompressed_path)



from abc import ABC, abstractmethod
from typing import Dict, Any, Union


class BlockInteriorProcessor(ABC):
    """
    Block interior processor (BIP) is a class that processes the interior of a block.
    """
    def __init__(self, read_depth: int = 2):
        """
        Create an explainer.
        :param read_depth: the depth of the tree to read (only that level and beyond will be read)
        """
        self.read_depth = read_depth

    @abstractmethod
    def parse(self, tag: str, meta: Dict[str, str], tree: Dict[str, Any]) -> Union[Dict[str, Any], None]:
        pass


class DefaultBIP(BlockInteriorProcessor):
    """
    Default BIP.
    """
    def parse(self, tag: str, meta: Dict[str, str], tree: Dict[str, Any]) -> Union[Dict[str, Any], None]:
        return tree



import os
from typing import List, Callable, Union
import zstandard as zstd
import py7zr
import shutil
import psutil

from twb.logger import twb_logger


def get_file_list(input_path: str) -> List[str]:
    """
    Get the list of files in the input directory.
    :param input_path: the input directory
    :raise FileNotFoundError: if the path does not exist
    :return: the list of files
    """
    # If the path does not exist, raise an error.
    if not os.path.exists(input_path):
        raise FileNotFoundError('The path does not exist.')

    # If the input path is a file, return the list with only the file path.
    if os.path.isfile(input_path):
        return [input_path]

    # If the input path is a directory, return the list of files in the directory.
    all_files = []
    for root, directories, files in os.walk(input_path):
        for file in files:
            file_path = os.path.join(root, file)
            all_files.append(file_path)

    # Remove duplicate files.
    all_files = sorted(list(set(all_files)))
    return all_files


def get_decompress_output_path(input_path: str, output_dir: str):
    """
    Get the output path of the decompressed file.
    :param input_path: the input path
    :param output_dir: the output directory
    :return: the output path
    """
    # Split the path into its components
    _, file_name = os.path.split(input_path)

    # Might want to reconsider the output file structure.
    if not file_name.endswith('.7z'):
        return os.path.join(output_dir, file_name)

    decompressed_file_name = file_name.rstrip('.7z')

    # Add the new directory to the beginning of the path
    return os.path.join(output_dir, decompressed_file_name)


def compress_zstd(input_path: str, output_path: str):
    """
    Compress the blocks into a Zstandard file.
    :param input_path: the input path
    :param output_path: the output path
    """
    # Compress the blocks.
    compressor = zstd.ZstdCompressor()
    with open(input_path, "rb") as ifh, open(output_path, "wb") as ofh:
        compressor.copy_stream(ifh, ofh)


def decompress_zstd(input_path: str, output_path: str):
    """
    Decompress the blocks from a Zstandard file.
    :param input_path: the input path
    :param output_path: the output path
    """
    # Decompress the blocks.
    decompressor = zstd.ZstdDecompressor()
    with open(input_path, "rb") as ifh, open(output_path, "wb") as ofh:
        decompressor.copy_stream(ifh, ofh)


def compute_total_available_space(output_dir: str) -> int:
    """
    Deprecated: Compute the total available space in the output directory.
    """
    total_available_space = shutil.disk_usage(output_dir).free

    # Display the total available space in GB.
    total_available_space_gb = total_available_space / 1024 / 1024 / 1024
    print('[Build] RDS space limitation (deprecated):', round(total_available_space_gb, 2), 'GB.')

    return total_available_space


def get_estimated_size(path: str) -> int:
    if path.endswith('.7z'):
        with py7zr.SevenZipFile(path, 'r') as z:
            space = z.archiveinfo().uncompressed
        return space * 2
    elif path.endswith('.zst'):
        with open(path, 'rb') as f:
            # Get the frame information for the compressed file
            space = zstd.frame_content_size(f.read(18))
            if space <= 0:
                return os.path.getsize(path) * 2
            else:
                return space * 2
    else:
        return os.path.getsize(path) * 2


def get_memory_consumption() -> int:
    process = psutil.Process(os.getpid())
    memory_usage_mb = process.memory_info().rss / 1024 / 1024
    return round(memory_usage_mb, 3)


def _rmtree_error_handler(func, path, exc_info):
    twb_logger.error(f"Error occurred while calling {func.__name__} on {path}")
    twb_logger.error(f"Error details: {exc_info}")

    # TODO: We might be able to attempt to resolve the issue based on exc_info and then retry the operation.


def cleanup_dir(path: str, onerror: Union[Callable, None] = _rmtree_error_handler):
    """
    Clean up the directory.
    :param path: the directory path
    :param onerror: the error handler
    """
    if os.path.exists(path):
        try:
            shutil.rmtree(path, onerror=onerror)
        except Exception as e:
            twb_logger.error(e)



import os
from typing import List, Union
from tqdm import tqdm
import requests
from bs4 import BeautifulSoup
from abc import ABC, abstractmethod
from multiprocessing import Pool


class Downloader:
    """
    A downloader for downloading files from the web.

    Attributes:
        profile: The download profile.
        url_batches: The batches of URLs to download.
        context: The context for the download profile.
        downloaded_files: The downloaded files.
        is_started: Whether the downloader is started.
        is_completed: Whether the downloader is completed.
    """
    def __init__(self):
        self.profile = None
        self.url_batches = []
        self.context = dict()
        self.downloaded_files = []
        self.is_started = False
        self.is_completed = False
        pass

    def will_download_wiki_history_dump(self, date: Union[str, None] = None):
        """
        Returns True if the downloader will download the wiki history dump for the given date.
        :param date: The date of the wiki history dump. Format: YYYYMMDD. (default: latest)
        :raises Exception: If the downloader is already downloading a profile.
        :return: True if the downloader will download the wiki history dump for the given date.
        """
        if self.profile is not None:
            raise Exception("One downloader can only download one profile at a time.")

        # Assign the date to the context only if it is not None.
        if date is not None:
            self.context["date"] = date

        # Assign the profile.
        self.profile = WikiHistoryDumpDownloadProfile()

        url_batches = self.profile.get_url_batches(self.context)
        url_batches = list(set(url_batches))  # Remove duplicates.

        # Check the filename in the URL to avoid overwriting.
        filenames = set()
        for url in url_batches:
            curr_filename = url.split('/')[-1]
            if curr_filename in filenames:
                url_batches.remove(url)
            else:
                filenames.add(curr_filename)

        self.url_batches = url_batches

        print(f'[Downloader] Find {len(self.url_batches)} files downloadable in this profile.')

        if len(self.url_batches) == 0:
            raise Warning("No URLs are found for the given date.")

    def start(self,
              output_dir: str,
              num_proc: Union[int, None] = None,
              limit: Union[int, None] = None):
        """
        Starts downloading the profile.
        :param output_dir: The output directory.
        :param num_proc: The number of processes to use for downloading. (default: number of CPUs)
                         (Might be limited by the profile.)
        :param limit: The maximum number of files to download. (default: no limit)
        """
        if self.profile is None:
            raise Exception("No profile is assigned to the downloader.")
        if self.url_batches is None or len(self.url_batches) == 0:
            raise Exception("No URLs are assigned to the downloader.")
        if self.is_completed:
            raise Exception("The downloader is already completed.")
        if self.is_started:
            raise Exception("The downloader is already started.")

        self.is_started = True

        url_batches = self.url_batches

        # Check the filename locally to avoid overwriting.
        for url in url_batches:
            target_path = os.path.join(output_dir, url.split('/')[-1])
            if os.path.exists(target_path):
                url_batches.remove(url)

        # Limit the number of files to download.
        if limit is not None:
            if limit <= 0:
                raise Warning('If you want to download all files, please set limit to None or do not define it.')
            url_batches = url_batches[:limit]

        if len(url_batches) == 0:
            print('[Downloader] No files need to be downloaded. Directly exit.')
            self.is_completed = True
            return

        # Create the output directory if it does not exist.
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        print(f'[Downloader] Will download {len(url_batches)} files.')

        if self.profile.max_processes is not None:
            max_num_proc = self.profile.max_processes
            if num_proc is not None and num_proc > max_num_proc:
                print(f'[Downloader] The number of processes is limited to {max_num_proc} by the profile.')
            num_proc = min(max_num_proc, num_proc if num_proc is not None else os.cpu_count())
        else:
            num_proc = num_proc if num_proc is not None else os.cpu_count()

        # Download the files.
        print(f'[Downloader] Start downloading. (# of assigned processes: {num_proc})')
        pool = Pool(processes=num_proc)
        pbar = tqdm(total=len(url_batches))
        downloaded_files = []

        def _callback_update_pbar(*args):
            nonlocal downloaded_files
            downloaded_files.append(args[0])
            pbar.update()

        # Start processes.
        for i in range(pbar.total):
            url = url_batches[i]
            pool.apply_async(download_executor, args=(url, output_dir), callback=_callback_update_pbar)

        # Wait for all processes to finish.
        pool.close()
        pool.join()

        pbar.close()

        self.downloaded_files = downloaded_files
        self.is_completed = True

        print(f'[Downloader] Downloading completed. {len(downloaded_files)} files are downloaded.')


class DownloadProfile(ABC):
    max_processes: Union[int, None] = None
    """
    The maximum number of processes to use for downloading.
    """

    @abstractmethod
    def get_url_batches(self, context: dict) -> List[str]:
        """
        Returns a list of batches of URLs to download.
        :return: A list of batches of URLs to download.
        """
        pass


# Download profile for the wiki history dump.
class WikiHistoryDumpDownloadProfile(DownloadProfile):
    max_processes = 3

    root_url = 'https://dumps.wikimedia.org'
    master_url = 'https://dumps.wikimedia.org/enwiki/'

    # Update this variable whenever the latest dump data is changed. Check `master_url` for the latest date.
    # Pick the date that is the closest to the current date AND must have the 7z history dump (some latest dates
    # do not have it).
    latest_dump_date = '20230301'

    def get_url_batches(self, context: dict) -> List[str]:
        urls = []
        context_date = context['date'] if 'date' in context else self.latest_dump_date
        url = self.master_url + context_date + '/'
        page = requests.get(url)
        soup = BeautifulSoup(page.content, "html.parser")
        results = soup.find_all('a', href=True)
        for f in results:
            href_link = f['href']
            if f'enwiki-{context_date}-pages-meta-history' in href_link and '7z' in href_link and 'xml' in href_link:
                urls.append(self.root_url + href_link)
        return urls


# TODO: Progress bar for downloading.
def download_executor(url: str, target_dir: str):
    # Compute the target path. Should assume that target dir exists.
    target_path = os.path.join(target_dir, url.split('/')[-1])
    # Download the file using a stream.
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(target_path, 'wb') as f:
            # Write the file in chunks of 10 MB for reduced memory usage.
            for chunk in r.iter_content(chunk_size=10 * 1024 * 1024):
                f.write(chunk)
    return target_path
